"""App Models."""

# Django
from django.db import models
from django.contrib.auth.models import Group
from django.core.validators import MaxValueValidator, MinValueValidator

# Third Party
from solo.models import SingletonModel

# Discord Obfuscate App
from discord_obfuscate.constants import ALLOWED_DIVIDERS, OBFUSCATION_METHODS


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        """Meta definitions"""

        managed = False
        default_permissions = ()
        permissions = (("basic_access", "Can access this app"),)


class DiscordRoleObfuscation(models.Model):
    """Configuration for obfuscating Discord roles tied to groups."""

    group = models.OneToOneField(
        Group,
        on_delete=models.CASCADE,
        related_name="discord_obfuscation",
    )
    opt_out = models.BooleanField(
        default=True,
        help_text="When enabled, the original group name is used.",
    )
    obfuscation_type = models.CharField(
        max_length=32,
        choices=[(key, label) for key, (label, _, _) in OBFUSCATION_METHODS.items()],
        default="sha256_base32",
    )
    obfuscation_format = models.CharField(
        max_length=120,
        blank=True,
        default="",
        help_text="Format string using {hash8}, {hash12}, {hash16}.",
    )
    divider_characters = models.CharField(
        max_length=64,
        blank=True,
        default="",
        help_text="Comma-separated list of divider characters.",
    )
    min_chars_before_divider = models.PositiveIntegerField(
        default=0,
        help_text="Minimum number of characters before a divider is used.",
    )
    custom_name = models.CharField(
        max_length=100,
        blank=True,
        default="",
        help_text="If set, overrides the obfuscation method.",
    )
    use_random_key = models.BooleanField(
        default=False,
        help_text=(
            "If enabled, obfuscation uses a random key instead of the group name. "
            "Rotation tasks only apply to entries with this enabled."
        ),
    )
    random_key = models.CharField(
        max_length=16,
        blank=True,
        default="",
        help_text="Random 16-character key used for obfuscation when enabled.",
    )
    random_key_rotate_name = models.BooleanField(
        default=True,
        help_text=(
            "Allow the rotation task to rename this role. "
            "Only applies when random key mode is enabled."
        ),
    )
    random_key_rotate_position = models.BooleanField(
        default=True,
        help_text=(
            "Allow the rotation task to reposition this role. "
            "Only applies when random key mode is enabled."
        ),
    )
    role_color = models.CharField(
        max_length=7,
        blank=True,
        default="",
        help_text="Optional role color in hex (#RRGGBB).",
    )
    role_id = models.BigIntegerField(
        null=True,
        blank=True,
        help_text="Cached Discord role ID for rename operations.",
    )
    last_obfuscated_name = models.CharField(
        max_length=100,
        blank=True,
        default="",
        verbose_name="Obfuscated Name",
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        verbose_name = "Discord Role Obfuscation"
        verbose_name_plural = "Discord Role Obfuscations"
        ordering = ["group__name"]

    def __str__(self):
        return self.group.name

    def get_dividers(self):
        return [d for d in self.divider_characters.split(",") if d]

    def set_dividers(self, values):
        unique = []
        for divider in values or []:
            if divider in ALLOWED_DIVIDERS and divider not in unique:
                unique.append(divider)
        self.divider_characters = ",".join(unique)


class DiscordObfuscateConfig(SingletonModel):
    """Global settings for Discord Obfuscate."""

    sync_on_save = models.BooleanField(
        default=True,
        help_text="Queue a role rename task when a config is saved in admin.",
    )
    default_opt_out = models.BooleanField(
        default=False,
        help_text="Default opt-out for newly created entries.",
    )
    default_use_random_key = models.BooleanField(
        default=False,
        help_text="Default random key setting for newly created entries.",
    )
    default_random_key_rotate_name = models.BooleanField(
        default=True,
        help_text="Default rotate-name setting for new random-key entries.",
    )
    default_random_key_rotate_position = models.BooleanField(
        default=True,
        help_text="Default rotate-position setting for new random-key entries.",
    )
    default_obfuscation_type = models.CharField(
        max_length=32,
        choices=[(key, label) for key, (label, _, _) in OBFUSCATION_METHODS.items()],
        default="sha256_base32",
        help_text="Default obfuscation type for new entries.",
    )
    default_divider_characters = models.CharField(
        max_length=64,
        blank=True,
        default="",
        help_text="Default divider characters for new entries.",
    )
    default_min_chars_before_divider = models.PositiveIntegerField(
        default=0,
        help_text="Default minimum chars before divider for new entries.",
    )
    random_key_rotation_enabled = models.BooleanField(
        default=False,
        help_text="Enable periodic rotation of random obfuscation keys.",
    )
    random_key_reposition_enabled = models.BooleanField(
        default=True,
        help_text="Enable automatic role repositioning during random key rotation.",
    )
    random_key_reposition_min_position = models.PositiveIntegerField(
        default=1,
        validators=[MinValueValidator(1), MaxValueValidator(249)],
        help_text=(
            "Lowest position for random-key role repositioning. "
            "1 is just above @everyone. Must leave enough space below position 250."
        ),
    )
    role_color_rule_sync_enabled = models.BooleanField(
        default=False,
        help_text="Enable periodic sync for role color rules.",
    )
    periodic_sync_enabled = models.BooleanField(
        default=False,
        help_text="Enable periodic full sync of roles via Celery beat.",
    )
    require_existing_role = models.BooleanField(
        default=True,
        help_text=(
            "Only use roles that already exist in Discord. "
            "When enabled, Alliance Auth will not be able to create new roles; "
            "missing roles must be created manually on the Discord server."
        ),
    )

    class Meta:
        verbose_name = "Discord Obfuscate Config"

    def __str__(self):
        return "Discord Obfuscate config"


class DiscordRoleOrderConfig(SingletonModel):
    """Manual role ordering settings."""

    ORDER_MODES = (
        ("desired", "Reorder to desired configuration at next reorder task"),
        ("shuffle", "Randomly reorganize unlocked roles"),
    )

    enabled = models.BooleanField(
        default=False,
        help_text="Enable manual role ordering during the rotation task.",
    )
    bot_role_id = models.BigIntegerField(
        null=True,
        blank=True,
        help_text="Discord role ID for the bot's highest role (locks roles above it).",
    )
    reorder_mode = models.CharField(
        max_length=16,
        choices=ORDER_MODES,
        default="shuffle",
        help_text="Controls how unlocked roles are positioned during the reorder task.",
    )
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        verbose_name = "Discord Role Order Config"

    def __str__(self):
        return "Discord Role Order config"


class DiscordRoleOrder(models.Model):
    """Stored manual ordering for Discord roles."""

    role_id = models.BigIntegerField(unique=True)
    role_name = models.CharField(max_length=100, blank=True, default="")
    role_color = models.CharField(max_length=7, blank=True, default="")
    sort_order = models.PositiveIntegerField(default=0)
    locked = models.BooleanField(default=False)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["sort_order", "role_name"]
        verbose_name = "Discord Role Order"
        verbose_name_plural = "Discord Role Orders"

    def __str__(self):
        return f"{self.role_name or self.role_id}"


class DiscordRoleColorRule(models.Model):
    """Rule for assigning random colors to matching roles."""

    name = models.CharField(
        max_length=100,
        help_text="Rule name shown in admin.",
    )
    pattern = models.CharField(
        max_length=150,
        help_text="Role name pattern. Use * as a wildcard.",
    )
    enabled = models.BooleanField(
        default=True,
        help_text="Enable automatic color assignment for this rule.",
    )
    case_sensitive = models.BooleanField(
        default=False,
        help_text="Match role names with case sensitivity.",
    )
    priority = models.PositiveIntegerField(
        default=100,
        help_text="Lower numbers run first when multiple rules match.",
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["priority", "name"]
        verbose_name = "Discord Role Color Rule"
        verbose_name_plural = "Discord Role Color Rules"

    def __str__(self):
        return f"{self.name}"


class DiscordRoleColorAssignment(models.Model):
    """Assigned color for a Discord role."""

    rule = models.ForeignKey(
        DiscordRoleColorRule,
        on_delete=models.CASCADE,
        related_name="assignments",
    )
    obfuscation = models.ForeignKey(
        DiscordRoleObfuscation,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="color_assignments",
    )
    role_id = models.BigIntegerField()
    role_name = models.CharField(max_length=100)
    color = models.CharField(max_length=7)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["rule", "role_name"]
        verbose_name = "Discord Role Color Assignment"
        verbose_name_plural = "Discord Role Color Assignments"
        constraints = [
            models.UniqueConstraint(fields=["role_id"], name="unique_role_color_role"),
            models.UniqueConstraint(fields=["color"], name="unique_role_color_value"),
        ]

    def __str__(self):
        return f"{self.role_name} ({self.color})"



# Generated by Discord Obfuscate on 2026-02-19

# Django
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (("basic_access", "Can access this app"),),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="DiscordRoleColorRule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Rule name shown in admin.",
                        max_length=100,
                    ),
                ),
                (
                    "pattern",
                    models.CharField(
                        help_text="Role name pattern. Use * as a wildcard.",
                        max_length=150,
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=True,
                        help_text="Enable automatic color assignment for this rule.",
                    ),
                ),
                (
                    "case_sensitive",
                    models.BooleanField(
                        default=False,
                        help_text="Match role names with case sensitivity.",
                    ),
                ),
                (
                    "priority",
                    models.PositiveIntegerField(
                        default=100,
                        help_text="Lower numbers run first when multiple rules match.",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Discord Role Color Rule",
                "verbose_name_plural": "Discord Role Color Rules",
                "ordering": ["priority", "name"],
            },
        ),
        migrations.CreateModel(
            name="DiscordObfuscateConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sync_on_save",
                    models.BooleanField(
                        default=True,
                        help_text="Queue a role rename task when a config is saved in admin.",
                    ),
                ),
                (
                    "default_opt_out",
                    models.BooleanField(
                        default=False,
                        help_text="Default opt-out for newly created entries.",
                    ),
                ),
                (
                    "default_use_random_key",
                    models.BooleanField(
                        default=False,
                        help_text="Default random key setting for newly created entries.",
                    ),
                ),
                (
                    "default_random_key_rotate_name",
                    models.BooleanField(
                        default=True,
                        help_text="Default rotate-name setting for new random-key entries.",
                    ),
                ),
                (
                    "default_random_key_rotate_position",
                    models.BooleanField(
                        default=True,
                        help_text="Default rotate-position setting for new random-key entries.",
                    ),
                ),
                (
                    "default_obfuscation_type",
                    models.CharField(
                        choices=[
                            ("sha256_hex", "SHA256 + hex"),
                            ("sha256_base32", "SHA256 + base32"),
                            ("blake2s_hex", "BLAKE2s + hex"),
                            ("blake2s_base32", "BLAKE2s + base32"),
                        ],
                        default="sha256_base32",
                        help_text="Default obfuscation type for new entries.",
                        max_length=32,
                    ),
                ),
                (
                    "default_divider_characters",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Default divider characters for new entries.",
                        max_length=64,
                    ),
                ),
                (
                    "default_min_chars_before_divider",
                    models.PositiveIntegerField(
                        default=0,
                        help_text="Default minimum chars before divider for new entries.",
                    ),
                ),
                (
                    "random_key_rotation_enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Enable periodic rotation of random obfuscation keys.",
                    ),
                ),
                (
                    "role_color_rule_sync_enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Enable periodic sync for role color rules.",
                    ),
                ),
                (
                    "periodic_sync_enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Enable periodic full sync of roles via Celery beat.",
                    ),
                ),
            ],
            options={
                "verbose_name": "Discord Obfuscate Config",
            },
        ),
        migrations.CreateModel(
            name="DiscordRoleObfuscation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "opt_out",
                    models.BooleanField(
                        default=True,
                        help_text="When enabled, the original group name is used.",
                    ),
                ),
                (
                    "obfuscation_type",
                    models.CharField(
                        choices=[
                            ("sha256_hex", "SHA256 + hex"),
                            ("sha256_base32", "SHA256 + base32"),
                            ("blake2s_hex", "BLAKE2s + hex"),
                            ("blake2s_base32", "BLAKE2s + base32"),
                        ],
                        default="sha256_base32",
                        max_length=32,
                    ),
                ),
                (
                    "obfuscation_format",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Format string using {hash8}, {hash12}, {hash16}.",
                        max_length=120,
                    ),
                ),
                (
                    "divider_characters",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Comma-separated list of divider characters.",
                        max_length=64,
                    ),
                ),
                (
                    "min_chars_before_divider",
                    models.PositiveIntegerField(
                        default=0,
                        help_text="Minimum number of characters before a divider is used.",
                    ),
                ),
                (
                    "custom_name",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="If set, overrides the obfuscation method.",
                        max_length=100,
                    ),
                ),
                (
                    "use_random_key",
                    models.BooleanField(
                        default=False,
                        help_text=(
                            "If enabled, obfuscation uses a random key instead of the group name. "
                            "Rotation tasks only apply to entries with this enabled."
                        ),
                    ),
                ),
                (
                    "random_key",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Random 16-character key used for obfuscation when enabled.",
                        max_length=16,
                    ),
                ),
                (
                    "random_key_rotate_name",
                    models.BooleanField(
                        default=True,
                        help_text=(
                            "Allow the rotation task to rename this role. "
                            "Only applies when random key mode is enabled."
                        ),
                    ),
                ),
                (
                    "random_key_rotate_position",
                    models.BooleanField(
                        default=True,
                        help_text=(
                            "Allow the rotation task to reposition this role. "
                            "Only applies when random key mode is enabled."
                        ),
                    ),
                ),
                (
                    "role_color",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Optional role color in hex (#RRGGBB).",
                        max_length=7,
                    ),
                ),
                (
                    "role_id",
                    models.BigIntegerField(
                        blank=True,
                        help_text="Cached Discord role ID for rename operations.",
                        null=True,
                    ),
                ),
                (
                    "last_obfuscated_name",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=100,
                        verbose_name="Obfuscated Name",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "group",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="discord_obfuscation",
                        to="auth.group",
                    ),
                ),
            ],
            options={
                "verbose_name": "Discord Role Obfuscation",
                "verbose_name_plural": "Discord Role Obfuscations",
                "ordering": ["group__name"],
            },
        ),
        migrations.CreateModel(
            name="DiscordRoleColorAssignment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("role_id", models.BigIntegerField()),
                ("role_name", models.CharField(max_length=100)),
                ("color", models.CharField(max_length=7)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "rule",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="assignments",
                        to="discord_obfuscate.discordrolecolorrule",
                    ),
                ),
            ],
            options={
                "verbose_name": "Discord Role Color Assignment",
                "verbose_name_plural": "Discord Role Color Assignments",
                "ordering": ["rule", "role_name"],
            },
        ),
        migrations.AddConstraint(
            model_name="discordrolecolorassignment",
            constraint=models.UniqueConstraint(
                fields=("role_id",), name="unique_role_color_role"
            ),
        ),
        migrations.AddConstraint(
            model_name="discordrolecolorassignment",
            constraint=models.UniqueConstraint(
                fields=("color",), name="unique_role_color_value"
            ),
        ),
    ]



"""
Discord Obfuscate test
"""

# Django
from django.test import TestCase

# Discord Obfuscate App
from discord_obfuscate.obfuscation import obfuscate_name

class TestDiscordObfuscate(TestCase):
    """
    TestDiscordObfuscate
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Test setup
        :return:
        :rtype:
        """

        super().setUpClass()

    def test_discord_obfuscate(self):
        """
        Dummy test function
        :return:
        :rtype:
        """

        result = obfuscate_name(
            "Test Group",
            "sha256_hex",
            "secret",
            prefix="grp",
            format_str="{prefix}{hash8}",
        )
        self.assertTrue(result.startswith("grp"))
        self.assertEqual(len(result), len("grp") + 8)



"""Runtime config helpers."""

# Django
from django.apps import apps

# Discord Obfuscate App
from discord_obfuscate.constants import (
    ALLOWED_DIVIDERS,
    DEFAULT_OBFUSCATE_METHOD,
    DEFAULT_PERIODIC_SYNC_ENABLED,
    DEFAULT_SYNC_ON_SAVE,
    OBFUSCATION_METHODS,
)


def _get_config():
    try:
        if not apps.ready:
            return None
        model = apps.get_model("discord_obfuscate", "DiscordObfuscateConfig")
        if model is None:
            return None
        return model.get_solo()
    except Exception:
        return None


def _get_role_order_config():
    try:
        if not apps.ready:
            return None
        model = apps.get_model("discord_obfuscate", "DiscordRoleOrderConfig")
        if model is None:
            return None
        return model.get_solo()
    except Exception:
        return None


def sync_on_save_enabled() -> bool:
    config = _get_config()
    if config:
        return bool(config.sync_on_save)
    return bool(DEFAULT_SYNC_ON_SAVE)


def periodic_sync_enabled() -> bool:
    config = _get_config()
    if config:
        return bool(config.periodic_sync_enabled)
    return bool(DEFAULT_PERIODIC_SYNC_ENABLED)


def role_color_rule_sync_enabled() -> bool:
    config = _get_config()
    if config:
        return bool(config.role_color_rule_sync_enabled)
    return False


def random_key_rotation_enabled() -> bool:
    config = _get_config()
    if config:
        return bool(config.random_key_rotation_enabled)
    return False


def random_key_reposition_enabled() -> bool:
    config = _get_config()
    if config:
        return bool(config.random_key_reposition_enabled)
    return True


def random_key_reposition_min_position() -> int:
    config = _get_config()
    if config:
        return int(config.random_key_reposition_min_position or 1)
    return 1


def require_existing_role() -> bool:
    config = _get_config()
    if config:
        return bool(config.require_existing_role)
    return True


def role_ordering_enabled() -> bool:
    config = _get_role_order_config()
    if config:
        return bool(config.enabled)
    return False


def role_order_bot_role_id() -> int | None:
    config = _get_role_order_config()
    if config and config.bot_role_id:
        return int(config.bot_role_id)
    return None


def role_order_mode() -> str:
    config = _get_role_order_config()
    if config and config.reorder_mode:
        return str(config.reorder_mode)
    return "shuffle"


def default_obfuscation_values() -> dict:
    config = _get_config()
    if config:
        obfuscation_type = config.default_obfuscation_type
        if obfuscation_type not in OBFUSCATION_METHODS:
            obfuscation_type = DEFAULT_OBFUSCATE_METHOD
        dividers = [d for d in config.default_divider_characters.split(",") if d]
        dividers = [d for d in dividers if d in ALLOWED_DIVIDERS]
        return {
            "opt_out": bool(config.default_opt_out),
            "use_random_key": bool(config.default_use_random_key),
            "random_key_rotate_name": bool(config.default_random_key_rotate_name),
            "obfuscation_type": obfuscation_type,
            "divider_characters": ",".join(dividers),
            "min_chars_before_divider": int(
                config.default_min_chars_before_divider or 0
            ),
        }

    return {
        "opt_out": False,
        "use_random_key": False,
        "random_key_rotate_name": True,
        "obfuscation_type": DEFAULT_OBFUSCATE_METHOD,
        "divider_characters": "",
        "min_chars_before_divider": 0,
    }



# Generated by Discord Obfuscate on 2026-02-22

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("discord_obfuscate", "0004_role_color_assignment_obfuscation"),
    ]

    operations = [
        migrations.CreateModel(
            name="DiscordRoleOrderConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Enable manual role ordering during the rotation task.",
                    ),
                ),
                (
                    "bot_role_id",
                    models.BigIntegerField(
                        blank=True,
                        help_text=(
                            "Discord role ID for the bot's highest role "
                            "(locks roles above it)."
                        ),
                        null=True,
                    ),
                ),
                (
                    "reorder_mode",
                    models.CharField(
                        choices=[
                            ("desired", "Reorder to desired configuration at next reorder task"),
                            ("shuffle", "Randomly reorganize unlocked roles"),
                        ],
                        default="desired",
                        help_text=(
                            "Controls how unlocked roles are positioned during the "
                            "reorder task."
                        ),
                        max_length=16,
                    ),
                ),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Discord Role Order Config",
            },
        ),
        migrations.CreateModel(
            name="DiscordRoleOrder",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("role_id", models.BigIntegerField(unique=True)),
                ("role_name", models.CharField(blank=True, default="", max_length=100)),
                ("role_color", models.CharField(blank=True, default="", max_length=7)),
                ("sort_order", models.PositiveIntegerField(default=0)),
                ("locked", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Discord Role Order",
                "verbose_name_plural": "Discord Role Orders",
                "ordering": ["sort_order", "role_name"],
            },
        ),
    ]



"""Monkey patches for Alliance Auth Discord service."""

# Standard Library
import logging

# Discord Obfuscate App
from discord_obfuscate.obfuscation import (
    obfuscated_names_for_role_names,
    obfuscated_user_group_names,
)

logger = logging.getLogger(__name__)

_PATCHED = False


def patch_discord_user_group_names() -> None:
    """Patch Discord service to use obfuscated group names."""
    global _PATCHED
    if _PATCHED:
        return

    try:
        from allianceauth.services.modules.discord import core as discord_core
    except Exception:
        logger.info("Discord service not available; skipping patch")
        return

    original = discord_core._user_group_names
    logger.info("Patching Alliance Auth Discord _user_group_names")

    def _patched_user_group_names(user, state_name=None):
        try:
            original_names = original(user, state_name=state_name)
            return obfuscated_names_for_role_names(original_names)
        except Exception:
            logger.exception("Failed to obfuscate group names, falling back")
            return original(user, state_name=state_name)

    discord_core._user_group_names = _patched_user_group_names
    logger.info("Patched Alliance Auth Discord _user_group_names")
    _PATCHED = True



"""Obfuscation helpers."""

# Standard Library
import base64
import hashlib
import hmac
import itertools
import logging
import secrets
import string
import time
import json
from dataclasses import dataclass, field
from collections.abc import Mapping
from typing import Dict, Iterable, List, Optional

# Django
from django.contrib.auth.models import Group, User

# Alliance Auth
from allianceauth.services.modules.discord.discord_client.helpers import RolesSet

# Discord Obfuscate App
from discord_obfuscate.app_settings import DISCORD_OBFUSCATE_SECRET
from discord_obfuscate.config import require_existing_role
from discord_obfuscate.constants import (
    ALLOWED_DIVIDERS,
    DEFAULT_OBFUSCATE_ENABLED,
    DEFAULT_OBFUSCATE_FORMAT,
    DEFAULT_OBFUSCATE_METHOD,
    DEFAULT_OBFUSCATE_PREFIX,
    OBFUSCATION_METHODS,
    ROLE_NAME_MAX_LEN,
)
from discord_obfuscate.models import DiscordRoleObfuscation

logger = logging.getLogger(__name__)

RANDOM_KEY_CHARS = string.ascii_letters + string.digits


def generate_random_key(length: int = 16) -> str:
    return "".join(secrets.choice(RANDOM_KEY_CHARS) for _ in range(length))



@dataclass(frozen=True)
class RoleNameResolution:
    """Resolved role name for a group."""

    group: Optional[Group]
    desired_name: str
    used_name: Optional[str]
    matched_role_id: Optional[int]
    used_original: bool


@dataclass
class RawRole:
    """Fallback role model when the client returns stripped objects."""

    id: int
    name: str
    position: int = 0
    color: int = 0
    managed: bool = False
    raw: dict = field(default_factory=dict, repr=False)

    def __getattr__(self, item):
        if item in self.raw:
            return self.raw[item]
        raise AttributeError(f"{type(self).__name__} has no attribute '{item}'")


class SimpleRolesSet:
    """Minimal RolesSet replacement for raw role payloads."""

    def __init__(self, roles: list):
        self._roles = list(roles or [])
        self._by_name = {
            getattr(role, "name", None): role
            for role in self._roles
            if getattr(role, "name", None)
        }
        self._by_id = {
            getattr(role, "id", None): role
            for role in self._roles
            if getattr(role, "id", None) is not None
        }

    def __iter__(self):
        return iter(self._roles)

    def __len__(self):
        return len(self._roles)

    def role_by_name(self, name: str):
        return self._by_name.get(name)

    def role_by_id(self, role_id: int):
        return self._by_id.get(role_id)


def _method_info(method: str) -> tuple:
    method = method or DEFAULT_OBFUSCATE_METHOD
    if method not in OBFUSCATION_METHODS:
        method = DEFAULT_OBFUSCATE_METHOD
    _, algo, encoding = OBFUSCATION_METHODS[method]
    return algo, encoding


def _hash_bytes(name: str, secret: str, algo: str) -> bytes:
    secret_bytes = str(secret or "").encode("utf-8")
    name_bytes = str(name).encode("utf-8")
    digestmod = hashlib.sha256
    if algo == "blake2s":
        digestmod = hashlib.blake2s
    return hmac.new(secret_bytes, name_bytes, digestmod).digest()


def _encode_hash(hash_bytes: bytes, encoding: str) -> str:
    if encoding == "base32":
        return base64.b32encode(hash_bytes).decode("ascii").rstrip("=")
    return hash_bytes.hex()


def _apply_format(format_str: str, tokens: dict) -> str:
    result = format_str or DEFAULT_OBFUSCATE_FORMAT
    for key, value in tokens.items():
        result = result.replace(f"{{{key}}}", value)
    return result


def _sanitize_output(value: str, allowed_dividers: list) -> str:
    cleaned = []
    divider_set = set(allowed_dividers or [])
    for char in value:
        if char.isalnum() or char in divider_set:
            cleaned.append(char)
    return "".join(cleaned)


def _insert_dividers(value: str, dividers: list, min_chars: int) -> str:
    if not dividers or min_chars <= 0:
        return value
    chunks = [value[i : i + min_chars] for i in range(0, len(value), min_chars)]
    if len(chunks) <= 1:
        return value
    out = chunks[0]
    for divider, chunk in zip(itertools.cycle(dividers), chunks[1:]):
        out += divider + chunk
    return out


def obfuscate_name(
    name: str,
    method: str,
    secret: str,
    prefix: str = "",
    format_str: str = "",
    dividers: Optional[list] = None,
    min_chars_before_divider: int = 0,
) -> str:
    """Create a deterministic obfuscated name."""
    algo, encoding = _method_info(method)
    hash_str = _encode_hash(_hash_bytes(name, secret, algo), encoding)

    format_str = format_str or DEFAULT_OBFUSCATE_FORMAT
    prefix = prefix or ""
    if prefix and "{prefix}" not in format_str:
        format_str = "{prefix}" + format_str

    tokens = {
        "prefix": _sanitize_output(prefix, ALLOWED_DIVIDERS),
        "hash8": hash_str[:8],
        "hash12": hash_str[:12],
        "hash16": hash_str[:16],
    }
    value = _apply_format(format_str, tokens)
    value = _sanitize_output(value, dividers or [])
    value = _insert_dividers(value, dividers or [], min_chars_before_divider)
    value = _sanitize_output(value, dividers or [])
    return value[:ROLE_NAME_MAX_LEN]


def role_name_for_group(
    group: Group,
    config: Optional[DiscordRoleObfuscation],
) -> str:
    """Determine the desired role name for a group based on config."""
    if config is None:
        return group.name
    if config and config.opt_out:
        return group.name
    if config and config.custom_name:
        dividers = config.get_dividers()
        return _sanitize_output(str(config.custom_name), dividers)[:ROLE_NAME_MAX_LEN]
    method = config.obfuscation_type if config else DEFAULT_OBFUSCATE_METHOD
    format_str = config.obfuscation_format if config else DEFAULT_OBFUSCATE_FORMAT
    dividers = config.get_dividers() if config else []
    min_chars = config.min_chars_before_divider if config else 0
    input_name = group.name
    if config and config.use_random_key and config.random_key:
        input_name = config.random_key
    return obfuscate_name(
        input_name,
        method,
        DISCORD_OBFUSCATE_SECRET,
        DEFAULT_OBFUSCATE_PREFIX,
        format_str,
        dividers,
        min_chars,
    )


def _rate_limit_delay(exc) -> float:
    resets_in = getattr(exc, "resets_in", None)
    if resets_in is None and exc.args:
        try:
            resets_in = float(exc.args[0])
        except (TypeError, ValueError):
            resets_in = None
    if resets_in is None:
        return 5.0
    try:
        delay = float(resets_in)
    except (TypeError, ValueError):
        return 5.0
    if delay > 60:
        delay = delay / 1000.0
    return max(delay + 0.25, 0.5)


def _roles_have_position(roles: list) -> bool:
    for role in roles:
        if hasattr(role, "position"):
            return True
    return False


def _safe_int(value, default=0) -> int:
    try:
        if value is None:
            return default
        return int(value)
    except (TypeError, ValueError):
        return default


def _normalize_role_id(value) -> int:
    return _safe_int(value, default=0)


def _fetch_raw_roles(client, guild_id):
    try:
        return client._api_request(method="get", route=f"guilds/{guild_id}/roles")
    except Exception:
        logger.exception("Failed to fetch raw roles from Discord API")
        return None


def _raw_role_from_payload(payload: dict) -> RawRole:
    role_id = _normalize_role_id(payload.get("id"))
    name = payload.get("name") or ""
    position = _safe_int(payload.get("position"), default=0)
    color = _safe_int(payload.get("color"), default=0)
    managed = bool(payload.get("managed", False))
    return RawRole(
        id=role_id,
        name=name,
        position=position,
        color=color,
        managed=managed,
        raw=payload,
    )


def _normalize_raw_roles(raw_roles):
    if raw_roles is None:
        return None
    if isinstance(raw_roles, list):
        return raw_roles
    if isinstance(raw_roles, (bytes, bytearray)):
        try:
            raw_roles = raw_roles.decode("utf-8")
        except Exception:
            return None
    if isinstance(raw_roles, str):
        try:
            parsed = json.loads(raw_roles)
            if isinstance(parsed, list):
                return parsed
            raw_roles = parsed
        except Exception:
            return None
    if isinstance(raw_roles, Mapping):
        data = raw_roles.get("data")
        if isinstance(data, list):
            return data
    if hasattr(raw_roles, "json"):
        try:
            parsed = raw_roles.json()
            if isinstance(parsed, list):
                return parsed
        except Exception:
            return None
    return None


def fetch_roleset(use_cache: bool = True, max_attempts: int = 3) -> RolesSet:
    """Fetch roles for the configured guild as RolesSet."""
    try:
        from allianceauth.services.modules.discord.core import (
            default_bot_client,
            DISCORD_GUILD_ID,
        )
        from allianceauth.services.modules.discord.discord_client.exceptions import (
            DiscordRateLimitExhausted,
        )

        for attempt in range(1, max_attempts + 1):
            try:
                roles = default_bot_client.guild_roles(
                    guild_id=DISCORD_GUILD_ID, use_cache=use_cache
                )
                roles = list(roles) if roles else []
                if roles and not _roles_have_position(roles):
                    raw_roles = _normalize_raw_roles(
                        _fetch_raw_roles(default_bot_client, DISCORD_GUILD_ID)
                    )
                    if isinstance(raw_roles, list) and raw_roles:
                        raw_objects = [
                            _raw_role_from_payload(role)
                            for role in raw_roles
                            if isinstance(role, Mapping)
                        ]
                        if raw_objects:
                            return SimpleRolesSet(raw_objects)
                    logger.warning(
                        "Raw role fetch did not return a list; position data unavailable."
                    )
                    return SimpleRolesSet(roles)
                return RolesSet(roles)
            except DiscordRateLimitExhausted as exc:
                if attempt >= max_attempts:
                    raise
                delay = _rate_limit_delay(exc)
                logger.warning(
                    "Rate limit hit fetching roles; retrying in %.2fs (attempt %s/%s)",
                    delay,
                    attempt,
                    max_attempts,
                )
                time.sleep(delay)
    except Exception:
        logger.exception("Failed to fetch roles from Discord")
        return RolesSet([])
    return RolesSet([])


def resolve_group_role_name(
    group: Group,
    roleset: RolesSet,
    config: Optional[DiscordRoleObfuscation] = None,
) -> RoleNameResolution:
    """Resolve role name to be used for a group."""
    desired = role_name_for_group(group, config)
    desired_role = roleset.role_by_name(desired)
    if desired_role:
        if desired != group.name:
            logger.debug("Resolved group %s to existing obfuscated role %s", group.name, desired)
        return RoleNameResolution(
            group=group,
            desired_name=desired,
            used_name=desired,
            matched_role_id=desired_role.id,
            used_original=(desired == group.name),
        )

    original_role = roleset.role_by_name(group.name)
    if original_role:
        if desired != group.name:
            logger.debug(
                "Desired role %s missing; using original for group %s",
                desired,
                group.name,
            )
        return RoleNameResolution(
            group=group,
            desired_name=desired,
            used_name=group.name,
            matched_role_id=original_role.id,
            used_original=True,
        )

    if not require_existing_role():
        return RoleNameResolution(
            group=group,
            desired_name=desired,
            used_name=desired,
            matched_role_id=None,
            used_original=(desired == group.name),
        )

    return RoleNameResolution(
        group=group,
        desired_name=desired,
        used_name=None,
        matched_role_id=None,
        used_original=False,
    )


def obfuscated_user_group_names(
    user: User,
    state_name: Optional[str] = None,
) -> List[str]:
    """Return obfuscated role names for a user's groups."""
    groups = list(user.groups.all())
    return obfuscated_names_for_groups(groups)


def obfuscated_names_for_groups(groups: Iterable[Group]) -> List[str]:
    """Return obfuscated role names for a set of groups."""
    groups = list(groups)
    if not DEFAULT_OBFUSCATE_ENABLED:
        return [group.name for group in groups]

    roleset = _load_roleset_with_retry()
    if roleset is None:
        return [group.name for group in groups]

    role_names: List[str] = []
    for group in groups:
        config = DiscordRoleObfuscation.objects.filter(group=group).first()
        resolution = resolve_group_role_name(group, roleset, config=config)
        if resolution.used_name:
            if resolution.used_name != group.name:
                logger.debug(
                    "Obfuscate group %s -> %s",
                    group.name,
                    resolution.used_name,
                )
            else:
                logger.debug("Using original name for group %s", group.name)
            role_names.append(resolution.used_name)
        else:
            logger.debug(
                "Skipping group %s because no matching role exists in Discord",
                group.name,
            )

    return role_names


def obfuscated_names_for_role_names(role_names: Iterable[str]) -> List[str]:
    """Obfuscate matching group names within an arbitrary list of role names."""
    role_names = list(role_names)
    if not role_names:
        return []
    if not DEFAULT_OBFUSCATE_ENABLED:
        return role_names

    groups = list(Group.objects.filter(name__in=role_names))
    if not groups:
        return role_names

    roleset = _load_roleset_with_retry()
    if roleset is None:
        return role_names

    name_map: Dict[str, Optional[str]] = {}
    for group in groups:
        config = DiscordRoleObfuscation.objects.filter(group=group).first()
        resolution = resolve_group_role_name(group, roleset, config=config)
        name_map[group.name] = resolution.used_name

    output: List[str] = []
    for name in role_names:
        desired = name_map.get(name)
        if name in name_map:
            if desired:
                output.append(desired)
            continue
        output.append(name)
    return output


def _load_roleset_with_retry() -> Optional[RolesSet]:
    roleset = fetch_roleset(use_cache=True)
    if roleset and len(roleset):
        return roleset
    logger.warning("Roleset cache empty; refetching from Discord API.")
    roleset = None
    for attempt in range(1, 4):
        roleset = fetch_roleset(use_cache=False)
        if roleset and len(roleset):
            return roleset
        if attempt < 3:
            logger.warning(
                "Roleset still empty; retrying in 5s (attempt %s/3).",
                attempt,
            )
            time.sleep(5)
    logger.warning(
        "Roleset unavailable; returning original group names to avoid role creation."
    )
    return None


def get_group_configs(groups: Iterable[Group]) -> Dict[int, DiscordRoleObfuscation]:
    """Fetch configs keyed by group id."""
    configs = DiscordRoleObfuscation.objects.filter(group__in=groups)
    return {cfg.group_id: cfg for cfg in configs}



"""Helpers for role color assignment."""

# Standard Library
import colorsys
import random
from typing import Iterable, List, Set


PALETTE_SIZE = 250
PALETTE_SATURATION = 0.65
PALETTE_LIGHTNESS = 0.5


def _int_to_hex(value: int) -> str:
    return f"#{value:06x}"


def _hex_to_int(value: str) -> int | None:
    value = (value or "").strip()
    if value.startswith("#"):
        value = value[1:]
    if len(value) != 6:
        return None
    try:
        return int(value, 16)
    except ValueError:
        return None


def build_palette(count: int = PALETTE_SIZE) -> List[int]:
    colors: List[int] = []
    if count <= 0:
        return colors
    for idx in range(count):
        hue = idx / float(count)
        r, g, b = colorsys.hls_to_rgb(hue, PALETTE_LIGHTNESS, PALETTE_SATURATION)
        value = (int(r * 255) << 16) + (int(g * 255) << 8) + int(b * 255)
        colors.append(value)
    return colors


def select_random_color(available: Iterable[int]) -> int | None:
    pool = list(available)
    if not pool:
        return None
    return random.SystemRandom().choice(pool)


def available_colors(
    palette: Iterable[int],
    used_colors: Set[int],
) -> List[int]:
    return [color for color in palette if color not in used_colors]


def to_hex(value: int) -> str:
    return _int_to_hex(value)


def to_int(value: str) -> int | None:
    return _hex_to_int(value)



# Generated by Codex on 2026-02-23

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("discord_obfuscate", "0005_role_ordering"),
    ]

    operations = [
        migrations.AlterField(
            model_name="discordroleorderconfig",
            name="reorder_mode",
            field=models.CharField(
                choices=[
                    ("desired", "Reorder to desired configuration at next reorder task"),
                    ("shuffle", "Randomly reorganize unlocked roles"),
                ],
                default="shuffle",
                help_text="Controls how unlocked roles are positioned during the reorder task.",
                max_length=16,
            ),
        ),
    ]



"""Signal handlers for Discord Obfuscate."""

# Standard Library
import logging

# Django
from django.contrib.auth.models import Group
from django.db import transaction
from django.db.models.signals import post_save
from django.dispatch import receiver

# Discord Obfuscate App
from discord_obfuscate.config import default_obfuscation_values, role_color_rule_sync_enabled
from discord_obfuscate.constants import DEFAULT_OBFUSCATE_METHOD
from discord_obfuscate.models import DiscordRoleObfuscation
from discord_obfuscate.tasks import sync_role_color_rules

logger = logging.getLogger(__name__)


@receiver(post_save, sender=Group)
def schedule_role_color_sync(sender, instance: Group, created: bool, **kwargs):
    """Queue role color sync when new groups are created."""
    if not created:
        return

    group_id = instance.pk

    def _after_commit():
        group = Group.objects.filter(pk=group_id).first()
        if not group:
            logger.warning("Group %s no longer exists; skipping obfuscation setup.", group_id)
            return
        defaults = default_obfuscation_values()
        defaults.setdefault("obfuscation_type", DEFAULT_OBFUSCATE_METHOD)
        defaults["opt_out"] = True
        obfuscation, created_cfg = DiscordRoleObfuscation.objects.get_or_create(
            group=group, defaults=defaults
        )
        if created_cfg:
            logger.info("Created obfuscation config for group %s.", group.name)
        else:
            logger.debug("Obfuscation config already exists for group %s.", group.name)

        if role_color_rule_sync_enabled():
            logger.info(
                "Scheduling role color sync after 30s for group %s.", group.name
            )
            sync_role_color_rules.apply_async(countdown=30)

    transaction.on_commit(_after_commit)



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Discord Obfuscate App
from discord_obfuscate import __version__


class DiscordObfuscateConfig(AppConfig):
    """App Config"""

    name = "discord_obfuscate"
    label = "discord_obfuscate"
    verbose_name = f"Discord Obfuscate v{__version__}"

    def ready(self):
        from discord_obfuscate.patches import patch_discord_user_group_names
        from discord_obfuscate import signals  # noqa: F401

        patch_discord_user_group_names()



"""Create or enable periodic tasks for discord_obfuscate."""

# Django
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

# Discord Obfuscate App


class Command(BaseCommand):
    help = "Create or enable periodic Celery beat tasks for discord_obfuscate."

    def handle(self, *args, **options):
        try:
            from django_celery_beat.models import CrontabSchedule, PeriodicTask
        except Exception as exc:
            raise CommandError(
                "django_celery_beat is required to create periodic tasks."
            ) from exc

        hourly = {
            "minute": "0",
            "hour": "*/1",
            "day_of_week": "*",
            "day_of_month": "*",
            "month_of_year": "*",
            "timezone": settings.TIME_ZONE,
        }
        every_three_days = {
            "minute": "0",
            "hour": "0",
            "day_of_week": "*",
            "day_of_month": "*/3",
            "month_of_year": "*",
            "timezone": settings.TIME_ZONE,
        }

        self._ensure_periodic_task(
            PeriodicTask,
            CrontabSchedule,
            "Obfuscate Discord: Sync all roles",
            "discord_obfuscate.tasks.periodic_sync_all_roles",
            hourly,
        )
        self._ensure_periodic_task(
            PeriodicTask,
            CrontabSchedule,
            "Obfuscate Discord: Rotate random keys",
            "discord_obfuscate.tasks.periodic_rotate_random_keys",
            every_three_days,
        )
        self._ensure_periodic_task(
            PeriodicTask,
            CrontabSchedule,
            "Obfuscate Discord: Sync role colors",
            "discord_obfuscate.tasks.periodic_sync_role_colors",
            hourly,
        )

    def _ensure_periodic_task(
        self,
        PeriodicTask,
        CrontabSchedule,
        name: str,
        task_path: str,
        schedule: dict,
    ):
        minute = schedule.get("minute", "0")
        hour = schedule.get("hour", "*/1")
        day_of_week = schedule.get("day_of_week", "*")
        day_of_month = schedule.get("day_of_month", "*")
        month_of_year = schedule.get("month_of_year", "*")
        timezone = schedule.get("timezone")

        task = PeriodicTask.objects.filter(name=name).first()
        if task:
            update_fields = []
            if task.task != task_path:
                task.task = task_path
                update_fields.append("task")
            if not task.enabled:
                task.enabled = True
                update_fields.append("enabled")
            if update_fields:
                task.save(update_fields=update_fields)
                self.stdout.write(f"Updated periodic task '{name}'.")
            else:
                self.stdout.write(f"Periodic task '{name}' already exists.")
            return

        crontab, _ = CrontabSchedule.objects.get_or_create(
            minute=minute,
            hour=hour,
            day_of_week=day_of_week,
            day_of_month=day_of_month,
            month_of_year=month_of_year,
            timezone=timezone,
        )

        task = PeriodicTask.objects.create(
            name=name,
            task=task_path,
            crontab=crontab,
            enabled=True,
        )

        self.stdout.write(
            f"Created periodic task '{task.name}' with schedule {minute} {hour} "
            f"{day_of_month} {month_of_year} {day_of_week} ({timezone})."
        )



"""Admin models."""

# Standard Library
import json
from collections.abc import Mapping

# Django
from django.contrib import admin, messages
from django.contrib.auth.models import Group
from django.http import JsonResponse
from django.urls import path

# Third Party
from solo.admin import SingletonModelAdmin

# Discord Obfuscate App
from discord_obfuscate.constants import DEFAULT_OBFUSCATE_METHOD
from discord_obfuscate.config import default_obfuscation_values, sync_on_save_enabled
from discord_obfuscate.forms import (
    DiscordObfuscateConfigForm,
    DiscordRoleObfuscationForm,
    DiscordRoleOrderConfigForm,
)
from discord_obfuscate.models import (
    DiscordObfuscateConfig,
    DiscordRoleColorAssignment,
    DiscordRoleColorRule,
    DiscordRoleObfuscation,
    DiscordRoleOrder,
    DiscordRoleOrderConfig,
)
from discord_obfuscate.obfuscation import (
    fetch_roleset,
    generate_random_key,
    role_name_for_group,
)
from discord_obfuscate.role_colors import to_hex
from discord_obfuscate.tasks import sync_all_roles, sync_group_role

# Register your models here.

def _role_position(role, default=0):
    if role is None:
        return default
    if isinstance(role, Mapping):
        value = role.get("position", default)
    else:
        value = getattr(role, "position", default)
    try:
        if value is None:
            return default
        return int(value)
    except (TypeError, ValueError):
        return default


def _role_sort_key(role):
    position = _role_position(role, default=0)
    role_id = getattr(role, "id", 0) or 0
    try:
        role_id = int(role_id)
    except (TypeError, ValueError):
        role_id = 0
    return (-position, role_id)


def _role_is_everyone(role) -> bool:
    if role is None:
        return False
    if isinstance(role, Mapping):
        return role.get("name") == "@everyone"
    return getattr(role, "name", "") == "@everyone"


@admin.register(DiscordRoleObfuscation)
class DiscordRoleObfuscationAdmin(admin.ModelAdmin):
    form = DiscordRoleObfuscationForm
    list_display = (
        "group",
        "role_exists",
        "opt_out",
        "use_random_key",
        "random_key_rotate_name",
        "obfuscation_type",
        "custom_name",
        "last_obfuscated_name",
    )
    search_fields = ("group__name", "custom_name")
    list_filter = ("opt_out", "obfuscation_type")
    actions = [
        "discover_roles",
        "toggle_opt_out",
        "sync_selected_roles",
        "sync_all_roles_action",
    ]
    fields = (
        "group",
        "opt_out",
        "custom_name",
        "use_random_key",
        "random_key",
        "random_key_rotate_name",
        "role_color_enabled",
        "role_color",
        "obfuscation_type",
        "obfuscation_format",
        "divider_characters",
        "min_chars_before_divider",
        "preview",
    )

    class Media:
        js = ("discord_obfuscate/admin_preview.js",)

    def get_queryset(self, request):
        qs = super().get_queryset(request).select_related("group")
        self._roleset = fetch_roleset(use_cache=True)
        return qs

    def role_exists(self, obj):
        roleset = getattr(self, "_roleset", None) or fetch_roleset(use_cache=True)
        desired = role_name_for_group(obj.group, obj)
        if roleset.role_by_name(desired):
            return True
        if roleset.role_by_name(obj.group.name):
            return True
        if obj.role_id:
            for role in roleset:
                if role.id == obj.role_id:
                    return True
        return False

    role_exists.boolean = True
    role_exists.short_description = "Role Exists"

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj=obj, **kwargs)
        roleset = fetch_roleset(use_cache=True)
        role_names = {role.name for role in roleset}
        qs = Group.objects.filter(name__in=role_names)
        if obj and obj.group_id:
            qs = (qs | Group.objects.filter(pk=obj.group_id)).distinct()
        if "group" in form.base_fields:
            form.base_fields["group"].queryset = qs
        return form

    @admin.action(description="Discover groups from Discord roles")
    def discover_roles(self, request, queryset):
        roleset = fetch_roleset(use_cache=True)
        role_names = {role.name for role in roleset}
        groups = Group.objects.filter(name__in=role_names)
        defaults = default_obfuscation_values()
        defaults.setdefault("obfuscation_type", DEFAULT_OBFUSCATE_METHOD)
        created = 0
        for group in groups:
            _, was_created = DiscordRoleObfuscation.objects.get_or_create(
                group=group,
                defaults=defaults,
            )
            if was_created:
                created += 1
        messages.success(request, f"Discovered {created} new groups.")

    @admin.action(description="Toggle opt-out for selected roles")
    def toggle_opt_out(self, request, queryset):
        toggled = 0
        for config in queryset:
            config.opt_out = not config.opt_out
            config.save(update_fields=["opt_out", "updated_at"])
            if sync_on_save_enabled():
                sync_group_role.delay(config.group_id)
            toggled += 1
        if toggled:
            messages.success(
                request,
                f"Toggled opt-out for {toggled} groups.",
            )

    @admin.action(description="Sync selected roles now")
    def sync_selected_roles(self, request, queryset):
        count = 0
        for config in queryset:
            sync_group_role.delay(config.group_id)
            count += 1
        messages.success(request, f"Queued sync for {count} groups.")

    @admin.action(description="Sync all roles now")
    def sync_all_roles_action(self, request, queryset):
        sync_all_roles.delay()
        messages.success(request, "Queued sync for all groups.")

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not sync_on_save_enabled():
            return
        if form and form.has_changed():
            sync_group_role.delay(obj.group_id)

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "preview/",
                self.admin_site.admin_view(self.preview_view),
                name="discord_obfuscate_preview",
            ),
        ]
        return custom_urls + urls

    def preview_view(self, request):
        if request.method != "POST":
            return JsonResponse({"error": "POST required"}, status=405)

        group_id = request.POST.get("group")
        group_name = ""
        if group_id:
            try:
                group_name = Group.objects.get(pk=group_id).name
            except Group.DoesNotExist:
                group_name = ""

        if not group_name:
            return JsonResponse({"preview": ""})

        custom_name = (request.POST.get("custom_name") or "").strip()
        use_random_key = request.POST.get("use_random_key") in {"on", "true", "1"}
        random_key = (request.POST.get("random_key") or "").strip()
        rotate_name = request.POST.get("random_key_rotate_name") in {"on", "true", "1"}
        opt_out = request.POST.get("opt_out") in {"on", "true", "1"}
        obfuscation_type = request.POST.get("obfuscation_type")
        obfuscation_format = (request.POST.get("obfuscation_format") or "").strip()
        min_chars = int(request.POST.get("min_chars_before_divider") or 0)
        dividers = request.POST.getlist("divider_characters")

        temp_group = Group(id=group_id, name=group_name)
        temp_config = DiscordRoleObfuscation(
            group=temp_group,
            opt_out=opt_out,
            obfuscation_type=obfuscation_type,
            obfuscation_format=obfuscation_format,
            custom_name=custom_name,
            min_chars_before_divider=min_chars,
            use_random_key=use_random_key,
            random_key=random_key or (generate_random_key(16) if use_random_key else ""),
            random_key_rotate_name=rotate_name if use_random_key else False,
        )
        temp_config.set_dividers(dividers)
        preview = role_name_for_group(temp_group, temp_config)

        return JsonResponse({"preview": preview})


@admin.register(DiscordObfuscateConfig)
class DiscordObfuscateConfigAdmin(SingletonModelAdmin):
    form = DiscordObfuscateConfigForm
    fieldsets = (
        (
            "Sync Behavior",
            {
                "description": (
                    "Controls when sync runs and whether roles must already exist in Discord."
                ),
                "fields": (
                    "sync_on_save",
                    "periodic_sync_enabled",
                    "require_existing_role",
                )
            },
        ),
        (
            "Defaults for New Entries",
            {
                "description": "Applied when new per-group configs are created.",
                "fields": (
                    "default_opt_out",
                    "default_use_random_key",
                    "default_random_key_rotate_name",
                    "default_obfuscation_type",
                    "default_divider_characters",
                    "default_min_chars_before_divider",
                )
            },
        ),
        (
            "Random Key Rotation",
            {
                "description": "Controls periodic random key rotation.",
                "fields": (
                    "random_key_rotation_enabled",
                )
            },
        ),
        (
            "Role Color Rules",
            {
                "description": "Controls periodic color assignment for matching roles.",
                "fields": ("role_color_rule_sync_enabled",),
            },
        ),
    )


@admin.register(DiscordRoleOrderConfig)
class DiscordRoleOrderConfigAdmin(SingletonModelAdmin):
    form = DiscordRoleOrderConfigForm
    change_form_template = "admin/discord_obfuscate/discordroleorderconfig/change_form.html"
    fieldsets = (
        (
            "Role Ordering",
            {
                "description": (
                    "Configure manual role ordering and select the bot role "
                    "to lock roles above it."
                ),
                "fields": (
                    "enabled",
                    "bot_role_id",
                )
            },
        ),
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj=obj, **kwargs)
        roleset = fetch_roleset(use_cache=True)
        roles = sorted(list(roleset), key=_role_sort_key)
        choices = [("", "---------")]
        for role in roles:
            choices.append((str(role.id), f"{role.name} ({role.id})"))
        if "bot_role_id" in form.base_fields:
            form.base_fields["bot_role_id"].choices = choices
        return form

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not obj.enabled:
            return
        payload_raw = request.POST.get("role_order_data") or ""
        if not payload_raw:
            return
        try:
            payload = json.loads(payload_raw)
        except (TypeError, ValueError):
            messages.error(request, "Failed to parse role ordering payload.")
            return
        if not isinstance(payload, list):
            messages.error(request, "Invalid role ordering payload.")
            return

        roleset = fetch_roleset(use_cache=False)
        roles_by_id = {role.id: role for role in roleset}
        seen_ids: set[int] = set()
        for index, item in enumerate(payload, start=1):
            try:
                role_id = int(item.get("role_id"))
            except (TypeError, ValueError):
                continue
            user_locked = bool(item.get("locked"))
            role = roles_by_id.get(role_id)
            role_name = role.name if role else ""
            role_color = ""
            if role and getattr(role, "color", 0):
                role_color = to_hex(int(role.color))
            DiscordRoleOrder.objects.update_or_create(
                role_id=role_id,
                defaults={
                    "sort_order": index,
                    "locked": user_locked,
                    "role_name": role_name,
                    "role_color": role_color,
                },
            )
            seen_ids.add(role_id)

        if seen_ids:
            DiscordRoleOrder.objects.exclude(role_id__in=seen_ids).delete()
            messages.success(request, "Saved role ordering.")

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        extra_context = extra_context or {}
        roleset = fetch_roleset(use_cache=False)
        roles = list(roleset)
        roles_by_id = {role.id: role for role in roles}
        roles_by_name = {role.name: role for role in roles}
        configs = list(DiscordRoleObfuscation.objects.select_related("group").filter(opt_out=True))
        config_by_role_id = {}

        for config in configs:
            role_id = None
            if config.role_id and config.role_id in roles_by_id:
                role_id = config.role_id
            else:
                desired = role_name_for_group(config.group, config)
                for name in (desired, config.last_obfuscated_name, config.group.name):
                    if name and name in roles_by_name:
                        role_id = roles_by_name[name].id
                        break
            if role_id and role_id not in config_by_role_id:
                config_by_role_id[role_id] = config

        order_entries = list(DiscordRoleOrder.objects.all())
        order_by_id = {entry.role_id: entry for entry in order_entries}
        display_ids = [role.id for role in sorted(roles, key=_role_sort_key)]

        obj = self.get_object(request, object_id) if object_id else None
        if obj is None:
            obj = DiscordRoleOrderConfig.get_solo()
        bot_role_id = getattr(obj, "bot_role_id", None)
        bot_role = roles_by_id.get(bot_role_id) if bot_role_id else None
        bot_position = _role_position(bot_role, default=None) if bot_role else None

        warnings = []
        if not roles:
            warnings.append("Failed to load roles from Discord.")
        else:
            positions = [_role_position(role, default=None) for role in roles]
            if len(roles) > 1 and (not positions or all(pos == 0 for pos in positions if pos is not None)):
                warnings.append(
                    "Role positions unavailable from Discord; ordering may be incorrect."
                )
        if bot_role_id and not bot_role:
            warnings.append("Configured bot role not found in Discord roles.")
        if not bot_role_id:
            warnings.append("Bot role not set; roles above the bot cannot be locked.")
        if obj and not obj.enabled:
            warnings.append("Role ordering is disabled. Enable it above to edit this table.")

        rows = []
        for role_id in display_ids:
            role = roles_by_id.get(role_id)
            if not role:
                continue
            config = config_by_role_id.get(role.id)
            reasons = []
            if _role_is_everyone(role):
                reasons.append("@everyone")
            if bot_role_id and role.id == bot_role_id:
                reasons.append("is bot")
            elif bot_position is not None and _role_position(role) >= bot_position:
                reasons.append("above bot")
            if config and config.opt_out:
                reasons.append("opt out")
            system_locked = bool(reasons)
            user_locked = bool(order_by_id.get(role.id).locked) if role.id in order_by_id else False
            color_value = ""
            if getattr(role, "color", 0):
                color_value = to_hex(int(role.color))
            rows.append(
                {
                    "role_id": role.id,
                    "name": role.name,
                    "color": color_value,
                    "position": _role_position(role, default=None),
                    "user_locked": user_locked,
                    "system_locked": system_locked,
                    "lock_reasons": ", ".join(reasons),
                }
            )

        extra_context["role_order_rows"] = rows
        extra_context["role_order_warnings"] = warnings
        extra_context["role_order_enabled"] = bool(getattr(obj, "enabled", False))
        return super().changeform_view(
            request,
            object_id=object_id,
            form_url=form_url,
            extra_context=extra_context,
        )


class DiscordRoleColorAssignmentInline(admin.TabularInline):
    model = DiscordRoleColorAssignment
    extra = 0
    fields = ("role_name", "role_id", "color", "updated_at")
    readonly_fields = ("role_name", "role_id", "color", "updated_at")
    can_delete = False
    show_change_link = False


@admin.register(DiscordRoleColorRule)
class DiscordRoleColorRuleAdmin(admin.ModelAdmin):
    list_display = ("name", "pattern", "enabled", "priority", "updated_at")
    list_filter = ("enabled",)
    search_fields = ("name", "pattern")
    fields = (
        "name",
        "pattern",
        "enabled",
        "case_sensitive",
        "priority",
    )
    inlines = [DiscordRoleColorAssignmentInline]


@admin.register(DiscordRoleColorAssignment)
class DiscordRoleColorAssignmentAdmin(admin.ModelAdmin):
    list_display = ("role_name", "color", "rule", "updated_at")
    list_filter = ("rule",)
    search_fields = ("role_name", "color")
    readonly_fields = ("rule", "role_name", "role_id", "color", "created_at", "updated_at")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False



"""Shared constants."""

ROLE_NAME_MAX_LEN = 100

DEFAULT_OBFUSCATE_ENABLED = True
DEFAULT_OBFUSCATE_METHOD = "sha256_base32"
DEFAULT_OBFUSCATE_FORMAT = "{hash12}"
DEFAULT_OBFUSCATE_PREFIX = ""
DEFAULT_SYNC_ON_SAVE = True
DEFAULT_PERIODIC_SYNC_ENABLED = False

ALLOWED_DIVIDERS = [
    "┃",
    "┇",
    "┆",
    "︲",
    "｜",
    "︱",
    "➖",
]

OBFUSCATION_METHODS = {
    "sha256_hex": ("SHA256 + hex", "sha256", "hex"),
    "sha256_base32": ("SHA256 + base32", "sha256", "base32"),
    "blake2s_hex": ("BLAKE2s + hex", "blake2s", "hex"),
    "blake2s_base32": ("BLAKE2s + base32", "blake2s", "base32"),
}



# Generated by Discord Obfuscate on 2026-02-20

# Django
from django.db import migrations, models
import django.db.models.deletion


def backfill_obfuscation(apps, schema_editor):
    Assignment = apps.get_model("discord_obfuscate", "DiscordRoleColorAssignment")
    Obfuscation = apps.get_model("discord_obfuscate", "DiscordRoleObfuscation")

    role_to_config = {
        row["role_id"]: row["id"]
        for row in Obfuscation.objects.exclude(role_id=None).values("id", "role_id")
    }
    if not role_to_config:
        return

    for assignment in Assignment.objects.filter(role_id__in=role_to_config.keys()):
        assignment.obfuscation_id = role_to_config.get(assignment.role_id)
        assignment.save(update_fields=["obfuscation"])


def clear_obfuscation(apps, schema_editor):
    Assignment = apps.get_model("discord_obfuscate", "DiscordRoleColorAssignment")
    Assignment.objects.update(obfuscation=None)


class Migration(migrations.Migration):
    dependencies = [
        ("discord_obfuscate", "0003_random_key_reposition_min_position"),
    ]

    operations = [
        migrations.AddField(
            model_name="discordrolecolorassignment",
            name="obfuscation",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="color_assignments",
                to="discord_obfuscate.discordroleobfuscation",
            ),
        ),
        migrations.RunPython(backfill_obfuscation, clear_obfuscation),
    ]



# Generated by Discord Obfuscate on 2026-02-20

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("discord_obfuscate", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="discordobfuscateconfig",
            name="random_key_reposition_enabled",
            field=models.BooleanField(
                default=True,
                help_text="Enable automatic role repositioning during random key rotation.",
            ),
        ),
    ]



"""Forms for Discord role obfuscation."""

# Standard Library
import re

# Django
from django import forms

# Discord Obfuscate App
from discord_obfuscate.constants import ALLOWED_DIVIDERS, OBFUSCATION_METHODS
from discord_obfuscate.config import default_obfuscation_values
from discord_obfuscate.models import (
    DiscordObfuscateConfig,
    DiscordRoleObfuscation,
    DiscordRoleOrderConfig,
)
from discord_obfuscate.obfuscation import (
    generate_random_key,
    role_name_for_group,
)

DIVIDER_CHOICES = [(char, char) for char in ALLOWED_DIVIDERS]
PLACEHOLDER_PATTERN = re.compile(r"\{hash8\}|\{hash12\}|\{hash16\}|\{prefix\}")


class DiscordRoleObfuscationForm(forms.ModelForm):
    """Form for editing obfuscation settings."""

    divider_characters = forms.MultipleChoiceField(
        required=False,
        choices=DIVIDER_CHOICES,
        widget=forms.CheckboxSelectMultiple,
        label="Dividers",
        help_text="Allowed separators for obfuscated output. Requires minimum characters per divider.",
    )
    preview = forms.CharField(
        required=False,
        disabled=True,
        label="Preview",
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
        help_text="Calculated output using the current settings.",
    )
    random_key = forms.CharField(
        required=False,
        label="Random key",
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
        help_text="Generated automatically when random key mode is enabled.",
    )
    role_color = forms.CharField(
        required=False,
        label="Role color",
        widget=forms.TextInput(attrs={"type": "color"}),
        help_text="Optional fixed role color in #RRGGBB format.",
    )
    role_color_enabled = forms.BooleanField(
        required=False,
        label="Apply role color",
        help_text="Uncheck to not send a color.",
    )

    class Meta:
        model = DiscordRoleObfuscation
        fields = [
            "group",
            "opt_out",
            "obfuscation_type",
            "obfuscation_format",
            "divider_characters",
            "min_chars_before_divider",
            "custom_name",
            "use_random_key",
            "random_key",
            "random_key_rotate_name",
            "role_color_enabled",
            "role_color",
            "preview",
        ]
        widgets = {
            "custom_name": forms.TextInput(attrs={"class": "form-control"}),
            "obfuscation_format": forms.TextInput(attrs={"class": "form-control"}),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.instance or not self.instance.pk:
            defaults = default_obfuscation_values()
            self.fields["opt_out"].initial = defaults.get("opt_out", False)
            self.fields["use_random_key"].initial = defaults.get("use_random_key", False)
            self.fields["random_key_rotate_name"].initial = defaults.get(
                "random_key_rotate_name", True
            )
            self.fields["obfuscation_type"].initial = defaults.get("obfuscation_type")
            divider_chars = defaults.get("divider_characters") or ""
            if divider_chars:
                self.fields["divider_characters"].initial = [
                    d for d in divider_chars.split(",") if d
                ]
            self.fields["min_chars_before_divider"].initial = defaults.get(
                "min_chars_before_divider", 0
            )
        if self.instance and self.instance.pk:
            self.fields["divider_characters"].initial = self.instance.get_dividers()
            self.fields["preview"].initial = role_name_for_group(
                self.instance.group, self.instance
            )
            self.fields["role_color_enabled"].initial = bool(self.instance.role_color)
        if "random_key" in self.fields:
            self.fields["random_key"].widget.attrs["readonly"] = "readonly"

    def clean_divider_characters(self):
        values = self.cleaned_data.get("divider_characters") or []
        return [val for val in values if val in ALLOWED_DIVIDERS]

    def clean_obfuscation_type(self):
        obfuscation_type = self.cleaned_data.get("obfuscation_type")
        if obfuscation_type not in OBFUSCATION_METHODS:
            return "sha256_base32"
        return obfuscation_type

    def clean_custom_name(self):
        custom_name = (self.cleaned_data.get("custom_name") or "").strip()
        dividers = set(self.cleaned_data.get("divider_characters") or [])
        if custom_name:
            for char in custom_name:
                if not (char.isalnum() or char in dividers):
                    raise forms.ValidationError(
                        "Custom name can only contain letters, numbers, and selected dividers."
                    )
        return custom_name

    def clean_obfuscation_format(self):
        value = (self.cleaned_data.get("obfuscation_format") or "").strip()
        dividers = set(self.cleaned_data.get("divider_characters") or [])
        stripped = PLACEHOLDER_PATTERN.sub("", value)
        for char in stripped:
            if not (char.isalnum() or char in dividers):
                raise forms.ValidationError(
                    "Format can only contain letters, numbers, placeholders, and selected dividers."
                )
        if value and not PLACEHOLDER_PATTERN.search(value):
            raise forms.ValidationError(
                "Format must include at least one of {hash8}, {hash12}, or {hash16}."
            )
        return value

    def clean_role_color(self):
        value = (self.cleaned_data.get("role_color") or "").strip()
        if not value:
            return ""
        if not value.startswith("#"):
            raise forms.ValidationError("Color must start with #.")
        if len(value) != 7:
            raise forms.ValidationError("Color must be in #RRGGBB format.")
        for char in value[1:]:
            if char.lower() not in "0123456789abcdef":
                raise forms.ValidationError("Color must be valid hex.")
        return value.lower()

    def clean_random_key(self):
        value = (self.cleaned_data.get("random_key") or "").strip()
        if not value:
            return ""
        if len(value) != 16:
            raise forms.ValidationError("Random key must be 16 characters.")
        if not value.isalnum():
            raise forms.ValidationError("Random key must be alphanumeric.")
        return value

    def clean(self):
        cleaned = super().clean()
        group = cleaned.get("group")
        if not group:
            self.add_error("group", "A group is required.")
        opt_out = cleaned.get("opt_out")
        if opt_out:
            cleaned["custom_name"] = ""
        use_random_key = cleaned.get("use_random_key")
        if use_random_key:
            cleaned["random_key"] = cleaned.get("random_key") or generate_random_key(16)
        else:
            cleaned["random_key"] = ""
            cleaned["random_key_rotate_name"] = False
        if not cleaned.get("role_color_enabled"):
            cleaned["role_color"] = ""
        dividers = cleaned.get("divider_characters") or []
        min_chars = cleaned.get("min_chars_before_divider") or 0
        if dividers and min_chars < 1:
            self.add_error(
                "min_chars_before_divider",
                "Minimum characters must be at least 1 when dividers are selected.",
            )
        return cleaned

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.set_dividers(self.cleaned_data.get("divider_characters") or [])
        if commit:
            instance.save()
        return instance


class DiscordObfuscateConfigForm(forms.ModelForm):
    """Form for global Discord obfuscation settings."""

    default_divider_characters = forms.MultipleChoiceField(
        required=False,
        choices=DIVIDER_CHOICES,
        widget=forms.CheckboxSelectMultiple,
        label="Default dividers",
        help_text="Default separators for new entries. Requires minimum characters per divider.",
    )

    class Meta:
        model = DiscordObfuscateConfig
        fields = [
            "sync_on_save",
            "default_opt_out",
            "default_use_random_key",
            "default_random_key_rotate_name",
            "default_obfuscation_type",
            "default_divider_characters",
            "default_min_chars_before_divider",
            "random_key_rotation_enabled",
            "role_color_rule_sync_enabled",
            "periodic_sync_enabled",
            "require_existing_role",
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        instance = getattr(self, "instance", None)
        if instance and instance.default_divider_characters:
            self.fields["default_divider_characters"].initial = [
                d
                for d in instance.default_divider_characters.split(",")
                if d in ALLOWED_DIVIDERS
            ]

    def clean_default_divider_characters(self):
        values = self.cleaned_data.get("default_divider_characters") or []
        return [val for val in values if val in ALLOWED_DIVIDERS]

    def save(self, commit=True):
        instance = super().save(commit=False)
        values = self.cleaned_data.get("default_divider_characters") or []
        instance.default_divider_characters = ",".join(
            [val for val in values if val in ALLOWED_DIVIDERS]
        )
        if commit:
            instance.save()
        return instance


class DiscordRoleOrderConfigForm(forms.ModelForm):
    """Form for role ordering settings."""

    bot_role_id = forms.ChoiceField(
        required=False,
        label="Bot role",
        help_text="Highest role of the bot; roles above it are locked.",
    )

    class Meta:
        model = DiscordRoleOrderConfig
        fields = [
            "enabled",
            "bot_role_id",
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.bot_role_id:
            self.initial["bot_role_id"] = str(self.instance.bot_role_id)

    def clean_bot_role_id(self):
        value = (self.cleaned_data.get("bot_role_id") or "").strip()
        if not value:
            return None
        try:
            return int(value)
        except (TypeError, ValueError):
            raise forms.ValidationError("Bot role must be a valid role ID.")



"""App Tasks"""

# Standard Library
import logging
import random
import time
from collections.abc import Mapping
from fnmatch import fnmatchcase

# Third Party
from celery import shared_task

from django.contrib.auth.models import Group

# Alliance Auth
# Discord Obfuscate App
from discord_obfuscate.constants import DEFAULT_OBFUSCATE_METHOD
from discord_obfuscate.config import (
    default_obfuscation_values,
    periodic_sync_enabled,
    random_key_rotation_enabled,
    role_ordering_enabled,
    role_order_bot_role_id,
    role_order_mode,
    role_color_rule_sync_enabled,
)
from discord_obfuscate.obfuscation import (
    fetch_roleset,
    generate_random_key,
    role_name_for_group,
)
from discord_obfuscate.role_colors import (
    available_colors,
    build_palette,
    select_random_color,
    to_hex,
    to_int,
)
from discord_obfuscate.models import (
    DiscordRoleColorAssignment,
    DiscordRoleColorRule,
    DiscordRoleObfuscation,
    DiscordRoleOrder,
)

logger = logging.getLogger(__name__)

# Create your tasks here


def _role_position(role, default=0):
    if role is None:
        return default
    if isinstance(role, Mapping):
        value = role.get("position", default)
    else:
        value = getattr(role, "position", default)
    try:
        if value is None:
            return default
        return int(value)
    except (TypeError, ValueError):
        return default


def _role_sort_key(role):
    position = _role_position(role, default=0)
    role_id = getattr(role, "id", 0) or 0
    try:
        role_id = int(role_id)
    except (TypeError, ValueError):
        role_id = 0
    return (-position, role_id)


def _role_is_everyone(role) -> bool:
    if role is None:
        return False
    if isinstance(role, Mapping):
        return role.get("name") == "@everyone"
    return getattr(role, "name", "") == "@everyone"


def _find_role_by_id(roleset, role_id):
    for role in roleset:
        if role.id == role_id:
            return role
    return None


def _update_role(role_id: int, name: str | None = None, color: int | None = None) -> bool:
    """Update a Discord role via bot client."""
    try:
        from allianceauth.services.modules.discord.core import (
            default_bot_client,
            DISCORD_GUILD_ID,
        )
        from allianceauth.services.modules.discord.discord_client.exceptions import (
            DiscordRateLimitExhausted,
        )

        route = f"guilds/{DISCORD_GUILD_ID}/roles/{role_id}"
        data = {}
        if name is not None:
            data["name"] = name
        if color is not None:
            data["color"] = color
        if not data:
            return True
        _api_request_with_retry(
            default_bot_client,
            DiscordRateLimitExhausted,
            method="patch",
            route=route,
            data=data,
        )
        default_bot_client._invalidate_guild_roles_cache(DISCORD_GUILD_ID)
        logger.info("Updated Discord role %s", role_id)
        return True
    except Exception:
        logger.exception("Failed to update role %s", role_id)
        return False


def _rename_role(role_id: int, new_name: str, color: int | None = None) -> bool:
    """Rename a Discord role via bot client."""
    return _update_role(role_id, name=new_name, color=color)


def _reorder_roles_payload(payload: list[dict]) -> bool:
    if not payload:
        return True
    try:
        from allianceauth.services.modules.discord.core import (
            default_bot_client,
            DISCORD_GUILD_ID,
        )
        from allianceauth.services.modules.discord.discord_client.exceptions import (
            DiscordRateLimitExhausted,
        )

        route = f"guilds/{DISCORD_GUILD_ID}/roles"
        _api_request_with_retry(
            default_bot_client,
            DiscordRateLimitExhausted,
            method="patch",
            route=route,
            data=payload,
        )
        default_bot_client._invalidate_guild_roles_cache(DISCORD_GUILD_ID)
        logger.info("Reordered %s roles via manual ordering", len(payload))
        return True
    except Exception:
        logger.exception("Failed to reorder roles via manual ordering")
        return False


def _build_manual_order_payload(
    roleset,
    bot_role_id: int | None,
    mode: str = "desired",
) -> list[dict]:
    roles = list(roleset)
    if not roles:
        return []

    roles_by_id = {role.id: role for role in roles}
    bot_role = roles_by_id.get(bot_role_id) if bot_role_id else None
    if bot_role_id and not bot_role:
        logger.warning("Manual role ordering enabled but bot role id %s not found.", bot_role_id)
        return []

    if bot_role_id is None:
        logger.warning("Manual role ordering enabled but bot role is not configured.")
        return []

    bot_position = _role_position(bot_role, default=None)
    if bot_position is None:
        logger.warning("Bot role position unavailable; skipping manual role ordering.")
        return []

    order_entries = list(DiscordRoleOrder.objects.all().order_by("sort_order", "role_name"))
    if not order_entries:
        logger.info("Manual role ordering enabled but no saved order exists.")
        return []

    system_locked_ids: set[int] = set()
    user_locked_ids: set[int] = {
        entry.role_id for entry in order_entries if entry.locked
    }
    opt_out_ids = _opt_out_role_ids(roleset)
    for role in roles:
        if _role_is_everyone(role):
            system_locked_ids.add(role.id)
            continue
        if _role_position(role) >= bot_position:
            system_locked_ids.add(role.id)
            continue
        if role.id in opt_out_ids:
            system_locked_ids.add(role.id)

    locked_ids = system_locked_ids | user_locked_ids
    movable_roles = [
        role
        for role in roles
        if _role_position(role) > 0 and _role_position(role) < bot_position
    ]
    movable_ids = {role.id for role in movable_roles}

    desired_ids = [
        entry.role_id
        for entry in order_entries
        if entry.role_id in movable_ids
    ]
    desired_set = set(desired_ids)
    remaining_ids = [
        role.id
        for role in sorted(movable_roles, key=_role_sort_key)
        if role.id not in desired_set
    ]
    ordered_ids = desired_ids + remaining_ids

    available_positions = sorted(
        {
            _role_position(role)
            for role in movable_roles
            if role.id not in locked_ids
        },
        reverse=True,
    )

    unlocked_ids = [role_id for role_id in ordered_ids if role_id not in locked_ids]
    if mode == "shuffle":
        shuffled = list(unlocked_ids)
        random.SystemRandom().shuffle(shuffled)
        unlocked_ids = shuffled

    payload = []
    for index, role_id in enumerate(unlocked_ids):
        if index >= len(available_positions):
            break
        payload.append({"id": role_id, "position": available_positions[index]})

    for role in movable_roles:
        if role.id in locked_ids:
            payload.append({"id": role.id, "position": _role_position(role)})

    return payload


def _opt_out_role_ids(roleset) -> set[int]:
    configs = list(
        DiscordRoleObfuscation.objects.select_related("group").filter(opt_out=True)
    )
    if not configs:
        return set()
    roles_by_id = {role.id: role for role in roleset}
    roles_by_name = {role.name: role for role in roleset}
    role_ids: set[int] = set()

    for config in configs:
        role_id = None
        if config.role_id and config.role_id in roles_by_id:
            role_id = config.role_id
        else:
            desired = role_name_for_group(config.group, config)
            for name in (desired, config.last_obfuscated_name, config.group.name):
                if name and name in roles_by_name:
                    role_id = roles_by_name[name].id
                    break
        if role_id:
            role_ids.add(role_id)

    return role_ids


def _api_request_with_retry(
    client,
    rate_limit_exc,
    method: str,
    route: str,
    data: dict | list,
    max_attempts: int = 3,
) -> None:
    for attempt in range(1, max_attempts + 1):
        try:
            client._api_request(method=method, route=route, data=data)
            return
        except rate_limit_exc as exc:
            if attempt >= max_attempts:
                raise
            delay = _rate_limit_delay(exc)
            logger.warning(
                "Rate limit hit; retrying in %.2fs (attempt %s/%s)",
                delay,
                attempt,
                max_attempts,
            )
            time.sleep(delay)


def _rate_limit_delay(exc) -> float:
    resets_in = getattr(exc, "resets_in", None)
    if resets_in is None and exc.args:
        try:
            resets_in = float(exc.args[0])
        except (TypeError, ValueError):
            resets_in = None
    if resets_in is None:
        return 5.0
    try:
        delay = float(resets_in)
    except (TypeError, ValueError):
        return 5.0
    if delay > 60:
        delay = delay / 1000.0
    return max(delay + 0.25, 0.5)


@shared_task
def sync_group_role(group_id: int) -> bool:
    """Sync role name for a single group."""
    try:
        group = Group.objects.get(pk=group_id)
    except Group.DoesNotExist:
        logger.warning("Group with id %s no longer exists", group_id)
        return False

    defaults = default_obfuscation_values()
    defaults.setdefault("obfuscation_type", DEFAULT_OBFUSCATE_METHOD)
    config, _ = DiscordRoleObfuscation.objects.get_or_create(group=group, defaults=defaults)
    return _sync_config(config)


def _sync_config(config: DiscordRoleObfuscation, roleset=None) -> bool:
    if config.use_random_key and not config.random_key:
        config.random_key = generate_random_key(16)
        config.save(update_fields=["random_key", "updated_at"])
    desired_name = role_name_for_group(config.group, config)
    logger.debug("Sync role for group %s -> %s", config.group.name, desired_name)
    color_value = None
    if config and config.role_color:
        value = config.role_color.strip()
        if value.startswith("#"):
            value = value[1:]
        if len(value) == 6:
            try:
                color_value = int(value, 16)
            except ValueError:
                color_value = None

    if roleset is None:
        roleset = fetch_roleset(use_cache=True)

    if not roleset or not len(roleset):
        if config.role_id and _rename_role(config.role_id, desired_name, color=color_value):
            config.last_obfuscated_name = desired_name
            config.save(update_fields=["last_obfuscated_name", "updated_at"])
            return True
        logger.info(
            "Skipping sync for group %s because roles could not be loaded",
            config.group.name,
        )
        return False

    desired_role = roleset.role_by_name(desired_name)
    if desired_role:
        config.role_id = desired_role.id
        config.last_obfuscated_name = desired_name
        config.save(update_fields=["role_id", "last_obfuscated_name", "updated_at"])
        logger.info("Role already matches desired name for group %s", config.group.name)
        if color_value is None:
            return True
        return _rename_role(desired_role.id, desired_name, color=color_value)

    role_to_rename = None
    if config.role_id:
        role_to_rename = _find_role_by_id(roleset, config.role_id)

    if not role_to_rename and config.last_obfuscated_name:
        role_to_rename = roleset.role_by_name(config.last_obfuscated_name)

    if not role_to_rename:
        role_to_rename = roleset.role_by_name(config.group.name)

    if not role_to_rename:
        logger.info("No matching role found for group %s", config.group.name)
        return False

    if role_to_rename.name == desired_name:
        config.role_id = role_to_rename.id
        config.last_obfuscated_name = desired_name
        config.save(update_fields=["role_id", "last_obfuscated_name", "updated_at"])
        logger.info("Role name already set for group %s", config.group.name)
        if color_value is None:
            return True
        return _rename_role(role_to_rename.id, desired_name, color=color_value)

    if _rename_role(role_to_rename.id, desired_name, color=color_value):
        config.role_id = role_to_rename.id
        config.last_obfuscated_name = desired_name
        config.save(update_fields=["role_id", "last_obfuscated_name", "updated_at"])
        return True

    return False


@shared_task
def sync_all_roles() -> int:
    """Sync role names for all groups with configs."""
    count = 0
    configs = list(DiscordRoleObfuscation.objects.select_related("group"))
    if configs:
        roleset = fetch_roleset(use_cache=False)
        for config in configs:
            if _sync_config(config, roleset=roleset):
                count += 1
        return count

    group_ids = list(Group.objects.values_list("id", flat=True))
    for group_id in group_ids:
        if sync_group_role(group_id):
            count += 1
    return count


def _role_name_matches(rule: DiscordRoleColorRule, role_name: str) -> bool:
    pattern = rule.pattern or ""
    if not pattern:
        return False
    if rule.case_sensitive:
        return fnmatchcase(role_name, pattern)
    return fnmatchcase(role_name.lower(), pattern.lower())


@shared_task
def sync_role_color_rules() -> int:
    """Assign colors to roles based on matching rules."""
    rules = list(
        DiscordRoleColorRule.objects.filter(enabled=True).order_by("priority", "id")
    )
    if not rules:
        return 0

    configs_with_roles = list(
        DiscordRoleObfuscation.objects.exclude(role_id=None).only("id", "role_id", "role_color")
    )
    pinned_role_ids = {
        cfg.role_id for cfg in configs_with_roles if cfg.role_color
    }
    obfuscation_by_role_id = {
        cfg.role_id: cfg for cfg in configs_with_roles if cfg.role_id
    }

    roleset = fetch_roleset(use_cache=False)
    roles_by_id = {role.id: role for role in roleset}

    existing_assignments = list(DiscordRoleColorAssignment.objects.all())
    stale_assignments = [
        assignment
        for assignment in existing_assignments
        if assignment.role_id not in roles_by_id
    ]
    if stale_assignments:
        DiscordRoleColorAssignment.objects.filter(
            id__in=[assignment.id for assignment in stale_assignments]
        ).delete()
        existing_assignments = [
            assignment
            for assignment in existing_assignments
            if assignment.role_id in roles_by_id
        ]

    assigned_role_ids = {assignment.role_id for assignment in existing_assignments}

    used_colors = set()
    for assignment in existing_assignments:
        value = to_int(assignment.color)
        if value is not None:
            used_colors.add(value)

    for role in roleset:
        color_value = getattr(role, "color", 0) or 0
        if color_value:
            used_colors.add(int(color_value))

    palette = build_palette()
    available = available_colors(palette, used_colors)
    created = 0

    for rule in rules:
        for role in roleset:
            if role.id in assigned_role_ids:
                continue
            if role.id in pinned_role_ids:
                continue
            if not _role_name_matches(rule, role.name):
                continue
            existing_color = getattr(role, "color", 0) or 0
            if existing_color:
                continue
            color_value = select_random_color(available)
            if color_value is None:
                logger.warning("No available colors left for rule %s", rule.name)
                return created
            if _update_role(role.id, color=color_value):
                DiscordRoleColorAssignment.objects.create(
                    rule=rule,
                    obfuscation=obfuscation_by_role_id.get(role.id),
                    role_id=role.id,
                    role_name=role.name,
                    color=to_hex(color_value),
                )
                assigned_role_ids.add(role.id)
                used_colors.add(color_value)
                available.remove(color_value)
                created += 1

    for assignment in existing_assignments:
        role = roles_by_id.get(assignment.role_id)
        if role and assignment.role_name != role.name:
            DiscordRoleColorAssignment.objects.filter(id=assignment.id).update(
                role_name=role.name
            )

    return created


@shared_task
def rotate_random_keys_and_reorder_roles() -> int:
    """Rotate random keys, sync role names, and reorder roles via role ordering config."""
    configs = list(
        DiscordRoleObfuscation.objects.select_related("group").filter(
            use_random_key=True
        )
    )
    if not configs and not role_ordering_enabled():
        return 0

    rename_targets = [config for config in configs if config.random_key_rotate_name]
    for config in rename_targets:
        config.random_key = generate_random_key(16)
        config.save(update_fields=["random_key", "updated_at"])

    updated = 0
    roleset = fetch_roleset(use_cache=False)
    for config in rename_targets:
        if _sync_config(config, roleset=roleset):
            updated += 1

    if role_ordering_enabled():
        bot_role_id = role_order_bot_role_id()
        payload = _build_manual_order_payload(roleset, bot_role_id, role_order_mode())
        if payload:
            _reorder_roles_payload(payload)
    return updated


@shared_task
def periodic_sync_all_roles() -> int:
    if not periodic_sync_enabled():
        return 0
    return sync_all_roles()


@shared_task
def periodic_sync_role_colors() -> int:
    if not role_color_rule_sync_enabled():
        return 0
    return sync_role_color_rules()


@shared_task
def periodic_rotate_random_keys() -> int:
    if not random_key_rotation_enabled():
        return 0
    return rotate_random_keys_and_reorder_roles()



# Generated by Discord Obfuscate on 2026-02-20

# Django
from django.db import migrations, models
import django.core.validators


class Migration(migrations.Migration):
    dependencies = [
        ("discord_obfuscate", "0002_random_key_reposition_enabled"),
    ]

    operations = [
        migrations.AddField(
            model_name="discordobfuscateconfig",
            name="random_key_reposition_min_position",
            field=models.PositiveIntegerField(
                default=1,
                help_text=(
                    "Lowest position for random-key role repositioning. "
                    "1 is just above @everyone. Must leave enough space below position 250."
                ),
                validators=[
                    django.core.validators.MinValueValidator(1),
                    django.core.validators.MaxValueValidator(249),
                ],
            ),
        ),
        migrations.AddField(
            model_name="discordobfuscateconfig",
            name="require_existing_role",
            field=models.BooleanField(
                default=True,
                help_text=(
                    "Only use roles that already exist in Discord. "
                    "When enabled, Alliance Auth will not be able to create new roles; "
                    "missing roles must be created manually on the Discord server."
                ),
            ),
        ),
    ]
